require 'ams_SoftenEdges.rb'

module AMS::SoftenEdges

  @opts = {
    :angle                      => 40,
    :smooth_normals             => true,
    :soften_coplanar            => true,
    :soften_curves              => true,
    :soften_material_dividers   => false,
    :recursive                  => true
  }
  @dialog = nil
  @initialized = false
  @op_started = false
  @border = nil

  class << self

    def soften_edges(ents, angle, smooth_normals = true, soften_coplanar = false, soften_curves = true, soften_material_dividers = true, recursive = true, depth = 0)
      model = ::Sketchup.active_model
      if recursive || depth == 0
        processed_definitions = {}
        ents.each { |e|
          if e.is_a?(::Sketchup::Group)
            soften_edges(e.entities, angle, smooth_normals, soften_coplanar, soften_curves, soften_material_dividers, recursive, depth + 1)
          elsif e.is_a?(::Sketchup::ComponentInstance)
            d = e.definition
            next if processed_definitions[d]
            processed_definitions[d] = true
            soften_edges(d.entities, angle, smooth_normals, soften_coplanar, soften_curves, soften_material_dividers, recursive, depth + 1)
          end
        }
      end
      edges = {}
      ents.each { |e|
        edges[e] = true if e.is_a?(::Sketchup::Edge)
      }
      edges.each { |e, s|
        #~ e.visible = true if e.hidden?
        efaces = e.faces
        if efaces.size != 2 || angle.zero? || (e.curve && !soften_curves)
          e.soft = false if e.soft?
          e.smooth = false if e.smooth?
          next
        end
        if efaces[0].material != efaces[1].material && !soften_material_dividers
          e.soft = false if e.soft?
          e.smooth = false if e.smooth?
          next
        end
        v1 = efaces[0].normal
        v2 = efaces[1].normal
        if v1.parallel?(v2) && !soften_coplanar
          e.soft = false if e.soft?
          e.smooth = false if e.smooth?
          next
        end
        theta = v1.angle_between(v2).radians
        if (theta < angle)
          e.soft = true unless e.soft?
          sn = smooth_normals ? true : false
          e.smooth = sn if e.smooth? != sn
        else
          e.soft = false if e.soft?
          e.smooth = false if e.smooth?
        end
      }
      nil
    end

    def soften_edges_from_selection
      ents_map = {}
      ::Sketchup.active_model.selection.each { |e|
        ents_map[e] = true
      }
      ents_map.each { |e, s|
        if e.is_a?(::Sketchup::Face)
          e.edges.each { |edge| ents_map[edge] = true }
        end
      }
      return false if ents_map.empty?
      soften_edges(ents_map.keys, @opts[:angle], @opts[:smooth_normals], @opts[:soften_coplanar], @opts[:soften_curves], @opts[:soften_material_dividers], @opts[:recursive])
      true
    end

    def reveal_quadrants(ents, smooth_normals = true, recursive = true, depth = 0)
      model = ::Sketchup.active_model
      if recursive || depth == 0
        processed_definitions = {}
        ents.each { |e|
          if e.is_a?(::Sketchup::Group)
            reveal_quadrants(e.entities, smooth_normals, recursive, depth + 1)
          elsif e.is_a?(::Sketchup::ComponentInstance)
            d = e.definition
            next if processed_definitions[d]
            processed_definitions[d] = true
            reveal_quadrants(d.entities, smooth_normals, recursive, depth + 1)
          end
        }
      end
      edges = {}
      ents.each { |e|
        edges[e] = true if e.is_a?(::Sketchup::Edge)
      }
      edges.each { |e, s|
        e.soft = false if e.soft?
        e.smooth = false if e.smooth?
      }
      edges.each { |e, s|
        e.start.edges.each { |edge1|
          next if edge1 == e
          vert = edge1.other_vertex(e.start)
          e.end.edges.each { |edge2|
            next if edge2 == e
            if edge2.other_vertex(e.end) == vert
              if e.length > edge1.length && e.length > edge2.length
                e.soft = true unless e.soft?
                e.smooth = true if smooth_normals && !e.smooth?
                edge1.soft = false if edge1.soft?
                edge1.smooth = false if edge1.smooth?
                edge2.soft = false if edge2.soft?
                edge2.smooth = false if edge2.smooth?
              end
            end
          }
        }
      }
    end

    def reveal_quadrants_from_selection
      ents_map = {}
      ::Sketchup.active_model.selection.each { |e|
        ents_map[e] = true
      }
      ents_map.each { |e, s|
        if e.is_a?(::Sketchup::Face)
          e.edges.each { |edge| ents_map[edge] = true }
        end
      }
      return false if ents_map.empty?
      reveal_quadrants(ents_map.keys, @opts[:smooth_normals], @opts[:recursive])
      true
    end

    def save_options
      ::Sketchup.write_default('Plugins_ams', 'Soften Edges', @opts.inspect.inspect[1..-2])
    end

    def load_options
      v = ::Sketchup.read_default('Plugins_ams', 'Soften Edges', '{}')
      begin
        res = eval(v)
        if res.is_a?(Hash) && res.size == @opts.size
          @opts.merge!(res)
        end
      rescue Exception => e
        # Do nothing; use default options.
      end
    end

    def options
      @opts.dup
    end

    def show_dialog(state)
      state = state ? true : false
      return false if state == dialog_visible?
      if state
        # Create a web dialog.
        iws = [280, 330]
        @dialog = ::UI::WebDialog.new('AMS Soften Edges', false, 'ams_SoftenEdges', iws.x, iws.y, 800, 600, false)
        @dialog.set_size(iws.x, iws.y) if @border.nil?
        # Attach callbacks
        @dialog.add_action_callback('init') { |acs, params|
          cmd = "$('#sel1').val(#{@opts[:angle]}).slider('refresh');"
          cmd << "$('#cb1').attr('checked', #{@opts[:smooth_normals]}).checkboxradio('refresh');"
          cmd << "$('#cb2').attr('checked', #{@opts[:soften_coplanar]}).checkboxradio('refresh');"
          cmd << "$('#cb3').attr('checked', #{@opts[:soften_curves]}).checkboxradio('refresh');"
          cmd << "$('#cb4').attr('checked', #{@opts[:recursive]}).checkboxradio('refresh');"
          cmd << "$('#cb5').attr('checked', #{@opts[:soften_material_dividers]}).checkboxradio('refresh');"
          @dialog.execute_script(cmd)
          next if @initialized
          @initialized = true
          ui_scale = Sketchup.version.to_i > 16 ? ::UI.scale_factor : 1.0
          data = eval(params)
          for i in 0..3
            data[i] *= ui_scale
          end
          @border = [iws.x - data[0], iws.y - data[1]] if @border.nil?
          @dialog.set_size(data[2] + @border.x, data[3] + @border.y)
        }
        @dialog.add_action_callback('window_focus') { |acs, params|
        }
        @dialog.add_action_callback('window_blur') { |acs, params|
          commit_operation
        }
        @dialog.add_action_callback('input_changed') { |acs, params|
          data = eval(params)
          case data[0]
            when :cb1
              @opts[:smooth_normals] = data[1]
            when :cb2
              @opts[:soften_coplanar] = data[1]
            when :cb3
              @opts[:soften_curves] = data[1]
            when :cb4
              @opts[:recursive] = data[1]
            when :cb5
              @opts[:soften_material_dividers] = data[1]
          end
          sel = ::Sketchup.active_model.selection
          next if sel.empty?
          start_operation
          soften_edges(sel, @opts[:angle], @opts[:smooth_normals], @opts[:soften_coplanar], @opts[:soften_curves], @opts[:soften_material_dividers], @opts[:recursive])
          commit_operation
          save_options
        }
        @dialog.add_action_callback('button_clicked') { |acs, params|
          start_operation
          case params
            when 'btn1'
              reveal_quadrants_from_selection
            when 'btn2'
              soften_edges_from_selection
          end
          commit_operation
        }
        @dialog.add_action_callback('slider_slide_start') { |acs, params|
          start_operation
        }
        @dialog.add_action_callback('slider_slide_stop') { |acs, params|
          commit_operation
          save_options
        }
        @dialog.add_action_callback('slider_slide') { |acs, params|
          use_new_op = !@op_started
          @opts[:angle] = params.to_f
          sel = ::Sketchup.active_model.selection.to_a
          next if sel.empty?
          start_operation if use_new_op
          soften_edges(sel, @opts[:angle], @opts[:smooth_normals], @opts[:soften_coplanar], @opts[:soften_curves], @opts[:soften_material_dividers], @opts[:recursive])
          commit_operation if use_new_op
          ::Sketchup.active_model.active_view.invalidate
        }
        # Set content
        dir = File.dirname(__FILE__)
        url = File.join(dir, 'soften_edges.html')
        @dialog.set_file(url)
        # Show dialog
        RUBY_PLATFORM =~ /mswin|mingw/i ? @dialog.show : @dialog.show_modal
        # Assign the on_close callback. Important: This must be called after
        # showing dialog in order to have it work on Mac OS X.
        @dialog.set_on_close() {
          @dialog = nil
          @initialized = false
          commit_operation
        }
      else
        @dialog.close
        commit_operation
      end
      true
    end

    def dialog_visible?
      @dialog ? true : false
    end

    def start_operation
      return if @op_started
      if ::Sketchup.version.to_i > 6
        ::Sketchup.active_model.start_operation('AMS Soften Edges', true)
      else
        ::Sketchup.active_model.start_operation('AMS Soften Edges')
      end
      @op_started = true
    end

    def commit_operation
      return unless @op_started
      ::Sketchup.active_model.commit_operation
      @op_started = false
    end

  end # class << self
end # module AMS::SoftenEdges

unless file_loaded?(__FILE__)
  AMS::SoftenEdges.load_options

  menu = UI.menu('Window')

  item = menu.add_item('AMS 增强柔化') {
    AMS::SoftenEdges.show_dialog( !AMS::SoftenEdges.dialog_visible? )
  }
  menu.set_validation_proc(item) {
    AMS::SoftenEdges.dialog_visible? ? MF_CHECKED : MF_UNCHECKED
  }

  UI.add_context_menu_handler do |context_menu|
    found = false
    ::Sketchup.active_model.selection.each { |e|
      if e.is_a?(Sketchup::Face) || e.is_a?(Sketchup::Group) || e.is_a?(Sketchup::ComponentInstance)
        found = true
        break
      end
    }
    if found
      context_menu.add_item('AMS 增强柔化') {
        AMS::SoftenEdges.show_dialog(true)
        AMS::SoftenEdges.start_operation
        AMS::SoftenEdges.soften_edges_from_selection
        AMS::SoftenEdges.commit_operation
      }
    end
  end

  file_loaded(__FILE__)
end
